from .main import *
from .spec_loader import *
from .transport import *

__version__ = "0.2.6"

__title__ = "decoda"
__description__ = "Decoder for J1939"
__url__ = "https://www.decoda.cc/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Andrew Dodd"
__email__ = "andrew.john.dodd@gmail.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) Andrew Dodd"
